#!/usr/bin/env python

############################################################################
#
# MODULE:       v.geobed 0.1 (alpha)
# AUTHOR(S):    Ivan Marchesini <ivan.marchesini AT irpi.cnr.it>
# PURPOSE:      Calculates geologic bedding attitudes starting from bedding traces and DEM
# COPYRIGHT:    (C) 2011 Ivan Marchesini
#
#               This program is free software under the GNU General Public
#               License (>=v2). Read the file COPYING that comes with GRASS
#               for details.
#
#############################################################################

#%module
#% description: calculate bedding attitude starding from bedding traces
#% keywords: bedding attitudes
#% keywords: geology, bedding attitude
#%end
#%option
#% key: bdt
#% type: string
#% gisprompt: old,vector,vector
#% description: name of the bedding traces map
#% required: yes
#%end
#%option
#% key: dtm
#% type: string
#% gisprompt: old,raster,dcell
#% description: name of the dtm
#% required: yes
#%end
#%option
#% key: sns
#% type: double
#% description: sinuosity threshold
#% required: yes
#% answer: 1.01
#%end
#%option
#% key: out
#% type: string
#% gisprompt: new,vector,vector
#% description: name of the bedding attitude layer
#% required: yes
#%end



import sys
import os
import atexit
import grass.script as grass
from grass.script import core as grasscore
import time
import datetime
from math import *


start=datetime.datetime.now()

def cleanup():
  pass

def main():
    
  bdt = options['bdt']
  dtm = options['dtm']
  sns = options['sns']
  out = options['out']

  #faccio partire il monitor
  mon = grass.read_command("d.mon", flags = 'p')
  #if mon.split("\n")[0] == "No monitor currently selected for output":
  if mon == '' :
    grass.run_command("d.mon",start="wx0")

  result = grass.find_file(name = "tmp", element = 'vector')
  if result['file']:
    grass.error("Please remove ore rename the existing \"tmp\" vector layer")
    sys.exit()

  #aggiungo il valore di sinuosita bidimensionale
  grass.message('---------------- I select only sinuous lines')
  vect=(bdt,"tmp")
  grass.run_command("g.copy", vect=vect)
  grass.run_command("v.db.addcolumn", map="tmp", layer=1, columns="sinuos double precision, startx double precision, starty double precision, endx double precision, endy double precision")
  grass.run_command("v.to.db", map="tmp", option="start", units="meters", columns="startx,starty")
  grass.run_command("v.to.db", map="tmp", option="end", units="meters", columns="endx,endy")
  where="endx=startx and endy=starty"
  grass.run_command("v.extract", input="tmp", output="tmp_close", type="line", where=where, overwrite=True)
  where="endx<>startx and endy<>starty"
  grass.run_command("v.extract", input="tmp", output="tmp_open", type="line", where=where, overwrite=True)  
  grass.run_command("v.to.db", map="tmp_open", option="sinuous", units="meters", columns="sinuos")
  #patch 26 settembre 2012
  linee_close=int(grass.vector_info_topo("tmp_close")["lines"])
  linee_open=int(grass.vector_info_topo("tmp_open")["lines"])
  if linee_close == 0 and linee_open > 0 :
    grass.run_command("g.copy", vect="tmp_open,tmp")
  elif linee_close > 0 and linee_open == 0 : 
    grass.run_command("g.copy", vect="tmp_close,tmp")
  elif linee_close > 0 and linee_open > 0 :
    grass.run_command("v.patch", input="tmp_close,tmp_open", output="tmp", flags="e", overwrite=True)
  else :
    grass.error("no lines selected... really strange.. exit")
    sys.exit()
  #grass.run_command("v.patch", input="tmp_close,tmp_open", output="tmp", flags="e", overwrite=True)
  #fine patch
  grass.run_command("g.remove", vect="tmp_open")
  grass.run_command("g.remove", vect="tmp_close")
  linee_iniziali=int(grass.vector_info_topo("tmp")["lines"])
  
  #drappeggio le linee sul DEM
  grass.message("---------------- I drape the beddings on the dtm")
  #resolution=float(grass.raster_info("dem")["ewres"])
  resolution=float(grass.raster_info(dtm)["ewres"])
  regionInc=resolution*2
  #take region values
  grass.run_command("g.region", vect="tmp", res=resolution)
  region = grasscore.region()
  E = region['e']
  W = region['w']
  N = region['n']
  S = region['s']
  grass.run_command("g.region", flags="a", e=E+regionInc, w=W-regionInc, s=S-regionInc, n=N+regionInc)
  grass.run_command("v.drape", input="tmp", type="line", elevation=dtm, output="bedd_3d", method="nearest", overwrite=True)
  #estraggo le sole linee che hanno sinuosita sopra il valore fornito o che hanno sinuosita nulla (linee chiuse)
  where="sinuos>"+sns+" or sinuos is null"
  grass.run_command("v.extract", input="bedd_3d", output="bed_trc_3d_sinuos", type="line", where=where, overwrite=True)
  #bdt3dInfo=grass.read_command("v.info", map="bed_trc_3d_sinuos", flags="t")
  #dictBdt3dInfo=grass.parse_key_val(bdt3dInfo, '=')
  #linee=int(dictBdt3dInfo['lines'])
  linee=int(grass.vector_info_topo("bed_trc_3d_sinuos")["lines"])
  lista = [linee, linee_iniziali]
  message = """
  --------------------------------------
  ----------- SELECTED %d LINES OF %d --
  --------------------------------------
  """
  grass.message(message % (lista[0], lista[1]))
  time.sleep(2)

  #inizializzo un contatore
  k=1
  categories=grass.read_command("v.category", flags="g", input="bed_trc_3d_sinuos", option="print", type="line")
  categories=categories.split("\n")
  categories.pop(-1) #remove last empy element of the list
  for i in categories:
	#estraggo la prima linea 
    grass.run_command("v.extract", input="bedd_3d", output="m_bacino"+str(i), cat=i, overwrite=True)	
		
    #calcolo le coordinate iniziali e finali di questa linea
    grass.run_command("g.region", vect="m_bacino"+str(i), res=resolution)
    region = grasscore.region()
    E = region['e']
    W = region['w']
    N = region['n']
    S = region['s']
    grass.run_command("g.region", flags="a", e=E+regionInc, w=W-regionInc, s=S-regionInc, n=N+regionInc)
    startc=grass.read_command("v.to.db", flags="p", map="m_bacino"+str(i), type="line", option="start", units="meters" , quiet=True)
    startx=float(startc.split("|")[1])
    starty=float(startc.split("|")[2])
    startz=float(startc.split("|")[3])
    endc=grass.read_command("v.to.db", flags="p", map="m_bacino"+str(i), type="line", option="end", units="meters", quiet=True)	
    endx=float(endc.split("|")[1])
    endy=float(endc.split("|")[2])
    endz=float(endc.split("|")[3])

    #creo un area chiusa costituita dal marker piu la linea che unisce gli estremi
    tmpfile = grass.tempfile()
    pfile = open(tmpfile, 'w')
    print>> pfile, "L", "2", "1"
    print>> pfile, startx, starty, startz
    print>> pfile, endx, endy, endz
    print>> pfile, "1", i
    pfile.close()
    grass.run_command("v.edit", input=tmpfile, flags="n", tool="add", map="m_bacino"+str(i), snap="node", thresh="1")
    grass.run_command("v.to.points", flags="i", input="m_bacino"+str(i), output="m_bacino_pnt"+str(i), type="line", llayer="1", use="vertex", dmax=resolution, overwrite=True)
    		
    #calcolo "l'ampiezza media" dell'area definita dal marker piu la linea che unisce gli estremi
    tmpfile = grass.tempfile()
    lfile = open(tmpfile, 'w')
    print>> lfile, "L", "2", "1"
    print>> lfile, startx, starty, startz
    print>> lfile, endx, endy, endz
    print>> lfile, "1", i
    lfile.close()
    grass.run_command("v.in.ascii", input=tmpfile, flags="nz", out="connect", format="standard", overwrite=True)
    grass.run_command("v.type", input="m_bacino"+str(i), output="m_bacino_bound"+str(i), from_type="line", to_type="boundary", overwrite=True)
    grass.run_command("v.to.3d", flags="r", input="m_bacino_bound"+str(i), output="m_bacino_bound_2d"+str(i), type="boundary", layer="1", overwrite=True)
    grass.run_command("v.clean", input="m_bacino_bound_2d"+str(i), output="m_bacino_bound_2d_break"+str(i), type="boundary", tool="break", overwrite=True) 
    grass.run_command("v.centroids", input="m_bacino_bound_2d_break"+str(i), output="m_bacino_bound_cent"+str(i), option="add", layer="1", cat="1", step="1", overwrite=True) 
    areaTot=grass.read_command("v.to.db", flags="psc", map="m_bacino_bound_cent"+str(i), type="centroid", option="area", units="meters", columns="m")
    dictAreaTot=grass.parse_key_val(areaTot, '|')
    #raw_input("Press ENTER to exit")
    area=round(float(dictAreaTot['total area']))
    baseTot=grass.read_command("v.to.db", flags="p", map="connect", type="line", option="length", units="meters", columns="p", quiet=True)
    base=round(float(baseTot.split("|")[1]))
    heigh=area/(base+1)

    #verifico se l'altezza media e' molto bassa (cioe' ho un marker molto allungato) o se l'altezza media e' molto alta (cioe' ho un marker che si piega su se stesso in modo molto stretto) 
    #e nel caso cambio la risoluzione in modo tale che il calcolo del piano inclinato avvenga anche per linee molto diritte
    heigh_min=round(resolution*3)
    heigh_max=round(area/(3*resolution))
    if heigh < heigh_min:
      res=heigh/5
      grass.run_command("g.region", res=res)
      grass.message("---------------- The resolution is: "+str(res))
    elif heigh > heigh_max and heigh != area:
      res=round(base/5)
      grass.run_command("g.region", res=res)
      grass.message("---------------- The resolution is: "+str(res))
    else:
      grass.message("---------------- The resolution is: "+str(resolution))
    #Produco la mappa del "piano" passante per il marker e le sue mappe derivate di slope ed aspect
    grass.run_command("v.extract", input="m_bacino_pnt"+str(i), output="m_pt_3dim"+str(i), layer="2", where="lcat="+str(i), overwrite=True)
    grass.run_command("v.to.rast", input="m_pt_3dim"+str(i), output="m_pt_3dim"+str(i), use="z", type="point", layer="2", value="1", rows="4096", overwrite=True)
    now=datetime.datetime.now()
    difftime=(now-start).seconds
    grass.message("----------PERCENT COMPLETED:")
    grass.core.percent(k-1,linee,1)
    lista = [k, linee, difftime, i]
    message = """
    -------------------------------------
    --------- USING LINE: %d OF %d ------
    --------- WORING TIME: %d seconds ---
    --------- LINE CATEGORY: %s ---------
    """
    grass.message(message % (lista[0], lista[1], lista[2], lista[3])) 
    grass.run_command("r.surf.nnbathy", input="m_pt_3dim"+str(i), output="m_pt_3dim_nn", alg="l", overwrite=True)
    grass.run_command("v.to.rast", input="m_bacino_bound_cent"+str(i), output="MASK", use="val", type="area", value="1", rows="4096", overwrite=True)
    grass.mapcalc("${out} = ${rast1}",
      out = "m_pt_3dim_nn"+str(i),
      rast1 = "m_pt_3dim_nn")
    grass.run_command("g.remove", rast="MASK")
    grass.run_command("d.erase")
    grass.run_command("d.rast", map="m_pt_3dim_nn"+str(i))
    grass.run_command("d.vect", map="m_bacino_bound_cent"+str(i), fcolor="none", width="2")
    grass.run_command("r.slope.aspect", elevation="m_pt_3dim_nn"+str(i), slope="slope"+str(i), aspect="aspect"+str(i), format="degrees", prec="FCELL", overwrite=True) 
    grass.mapcalc("${out} = ${rast1}/${rast1}",
      out = "m_pt_3dim_1_nn_poly",
      rast1 = "m_pt_3dim_nn"+str(i))
    grass.run_command("r.to.vect", input="m_pt_3dim_1_nn_poly", output="m_pt_3dim_1_nn_poly"+str(i), type="area", overwrite=True) 
    thresh=round((resolution**2)*2)
    grass.run_command("v.clean", input="m_pt_3dim_1_nn_poly"+str(i), output="m_pt_3dim_1_nn_poly_clean"+str(i), type="area", tool="rmarea", thresh=thresh, overwrite=True)
		
    #assegno le variabili slope ed azimuth al poligono che rappresenta il marker
    grass.run_command("g.copy", rast="m_pt_3dim_1_nn_poly,MASK")
    
    #assegno slope
    slopeUnivar=grass.read_command("r.univar", flags="eg", map="slope"+str(i))
    dictSlopeUnivar=grass.parse_key_val(slopeUnivar, '=')
    slope50=round(float(dictSlopeUnivar['median']))
    grass.run_command("g.remove", rast="MASK")
    columns=["slopestdev double precision", "slope50 double precision", "invslope50 double precision", "aspect double precision", "azimuth double precision", "circvar double precision", "angstddev double precision"]
    grass.run_command("v.db.addcolumn", map="m_pt_3dim_1_nn_poly_clean"+str(i), columns=columns)
    slopestdev=float(dictSlopeUnivar['stddev'])
    inv_slope50=90-slope50	
    grass.run_command("v.db.update", map="m_pt_3dim_1_nn_poly_clean"+str(i), column="slopestdev", value=slopestdev)
    grass.run_command("v.db.update", map="m_pt_3dim_1_nn_poly_clean"+str(i), column="slope50", value=slope50)
    grass.run_command("v.db.update", map="m_pt_3dim_1_nn_poly_clean"+str(i), column="invslope50", value=inv_slope50)
    
    #assegno azimuth
    grass.mapcalc("${out} = sin(${rast1})",
      out = "sen_aspect"+str(i),
      rast1 = "aspect"+str(i))
    grass.mapcalc("${out} = cos(${rast1})",
      out = "cos_aspect"+str(i),
      rast1 = "aspect"+str(i))		
    senUnivar=grass.read_command("r.univar", flags="eg", map="sen_aspect"+str(i))
    dictSenUnivar=grass.parse_key_val(senUnivar, '=')
    mean_sen=float(dictSenUnivar['mean'])
    cosUnivar=grass.read_command("r.univar", flags="eg", map="cos_aspect"+str(i))
    dictCosUnivar=grass.parse_key_val(cosUnivar, '=')
    mean_cos=float(dictCosUnivar['mean'])
    if mean_sen > 0 and mean_cos > 0:
      aspect_avg=round(atan(mean_sen/mean_cos)*360/6.28)
      azimuth=round(90-aspect_avg)
    elif mean_sen > 0 and mean_cos < 0: 
      aspect_avg=round(atan(mean_sen/mean_cos)*360/6.28+180)
      azimuth=round(360-aspect_avg+90)
    elif mean_sen < 0 and mean_cos < 0: 
      aspect_avg=round(atan(mean_sen/mean_cos)*360/6.28+180)
      azimuth=round(360-aspect_avg+90)
    elif mean_sen < 0 and mean_cos > 0: 
      aspect_avg=round(atan(mean_sen/mean_cos)*360/6.28+360)
      azimuth=round(-(aspect_avg-360)+90)
    grass.run_command("v.db.update", map="m_pt_3dim_1_nn_poly_clean"+str(i), column="aspect", value=aspect_avg)
    grass.run_command("v.db.update", map="m_pt_3dim_1_nn_poly_clean"+str(i), column="azimuth", value=azimuth)
    
    #assegno variabilita aspect
    sen_sum=float(dictSenUnivar['sum'])
    cos_sum=float(dictCosUnivar['sum'])
    n_cells=int(dictCosUnivar['n'])
    circvar=1-sqrt(sen_sum**2+cos_sum**2)/n_cells
    grass.run_command("v.db.update", map="m_pt_3dim_1_nn_poly_clean"+str(i), column="circvar", value=circvar)
		
	#assegno variabilita aspect, altro metodo
    grass.mapcalc("${out} = if(${rast1}<=180,${rast1},${rast1}-360)",
      out = "asp_semi"+str(i),
      rast1 = "aspect"+str(i))
    #r.mapcalc "asp_semi_$i=if(aspect_$i<=180,aspect_$i,aspect_$i-360)"
    if aspect_avg <= 180:
      aspect_avg_new=aspect_avg
    elif aspect_avg > 180:
      aspect_avg_new=aspect_avg-360
    grass.mapcalc("${out} = ${rast1}-${val2}",
      out = "diff_angoli"+str(i),
      rast1 = "asp_semi"+str(i),
      val2 = aspect_avg_new)
    grass.mapcalc("${out} = if(abs(${rast1})<=180,${rast1},(360-abs(${rast1})))",
      out = "diff_angoli_semi"+str(i),
      rast1 = "diff_angoli"+str(i))
    grass.mapcalc("${out} = ${rast1}^2",
      out = "diff_angoli_semi_squared"+str(i),
      rast1 = "diff_angoli_semi"+str(i))
    diffUnivar=grass.read_command("r.univar", flags="g", map="diff_angoli_semi_squared"+str(i))
    dictDiffUnivar=grass.parse_key_val(diffUnivar, '=')
    n_diffcell=float(dictDiffUnivar['n'])
    sum_diffcell=float(dictDiffUnivar['sum'])   
    angstddev=sqrt(sum_diffcell/n_diffcell)
    grass.run_command("v.db.update", map="m_pt_3dim_1_nn_poly_clean"+str(i), column="angstddev", value=angstddev)
		
    #trasformo i centroidi delle aree dei markers in punti, li sposto al centro della region e poi, laddove ne esista piu' di uno, prendo solo il primo
    grass.run_command("v.type", input="m_pt_3dim_1_nn_poly_clean"+str(i), output="ptbound"+str(i), from_type="centroid", to_type="point", overwrite=True)
    regionCent=grass.read_command("g.region", flags="cg", vect="m_pt_3dim_1_nn_poly_clean"+str(i))
    dictRegionCent=grass.parse_key_val(regionCent, '=')
    center_easting=float(dictRegionCent['center_easting'])
    center_northing=float(dictRegionCent['center_northing'])
    catList=grass.read_command("v.category", input="ptbound"+str(i), option="print", type="point")
    first_pt_cat=int(catList.split("\n")[0])
    grass.run_command("v.extract", input="ptbound"+str(i), output="pt"+str(i), type="point", cat=first_pt_cat, overwrite=True)
    grass.run_command("d.vect", map="pt"+str(i))
    raw_input("Press ENTER to exit")

    point_coor=grass.read_command("v.to.db", flags="p", map="pt"+str(i), type="point", option="coor", columns="x,y", quiet=True)
    point_easting=float(point_coor.split("|")[1])
    point_northing=float(point_coor.split("|")[2])
    x_move=center_easting-point_easting
    y_move=center_northing-point_northing
    point=(point_easting,point_northing)
    print(point)
    move=(x_move,y_move,0)
    grass.run_command("v.edit", tool="move", map="pt"+str(i), coords=point, move=move)
    grass.run_command("v.build", map="pt"+str(i))
    grass.run_command("d.vect", map="pt"+str(i), icon="geology/strike_circle", size="8")
    raw_input("Press ENTER to exit")


			
    #creo una mappa che contiene solo le giaciture medie per la zona di interesse
    if k == 1:
      grass.run_command("g.region", vect="pt"+str(i), flags="pg")
      region = grasscore.region()
      E = region['e']
      W = region['w']
      N = region['n']
      S = region['s']
      grass.run_command("g.remove", vect="giaciture_mediate")
      giaciture_mediate="giaciture_mediate"
      pt="pt"+str(i)
      vect=(pt,giaciture_mediate)
      grass.run_command("g.copy", vect=vect)
      grass.run_command("v.edit", tool="catdel", map=giaciture_mediate, ids="1-99999", cats="1-99999")
      bbox=(W,S,E,N)
      grass.run_command("v.edit", tool="catadd", map=giaciture_mediate, bbox=bbox, cats=i)
      grass.run_command("v.db.update", map=giaciture_mediate, column="cat", value=i)
      grass.message("---------------- Creating bedding attitude layer")
      #raw_input("Press ENTER to exit")
    else:
      grass.run_command("g.region", vect="pt"+str(i), flags="pg")
      region = grasscore.region()
      E = region['e']
      W = region['w']
      N = region['n']
      S = region['s']
      grass.run_command("v.patch", flags="ae", input="pt"+str(i), output="giaciture_mediate", overwrite=True)
      catList=grass.read_command("v.category", input="giaciture_mediate", option="print", type="point")
      ncat=int(catList.split("\n").__len__())-2
      last_cat=catList.split("\n")[ncat]
      grass.run_command("v.edit", tool="catdel", map="giaciture_mediate", ids="1-99999", cats=last_cat)
      bbox=(W,S,E,N)
      grass.run_command("v.edit", tool="catadd", map="giaciture_mediate", bbox=bbox, cat=i)
      where="cat="+last_cat
      grass.run_command("v.db.update", map="giaciture_mediate", column="cat", value=i, where=where)
      grass.message("---------------- patching bedding attitude layer")

    #rimuovo mappe temporanee
    grass.run_command("g.remove", vect="m_bacino_bound"+str(i))
    grass.run_command("g.remove", vect="m_bacino_bound_2d"+str(i))
    grass.run_command("g.remove", vect="m_bacino_bound_2d_break"+str(i))
    grass.run_command("g.remove", vect="m_bacino_bound_cent"+str(i))
    grass.run_command("g.remove", vect="m_bacino_pnt"+str(i))
    grass.run_command("g.remove", vect="m_pt_3dim_1_nn_poly"+str(i))
    grass.run_command("g.remove", vect="m_pt_3dim_1_nn_poly_clean"+str(i))	
    grass.run_command("g.remove", vect="m_bacino"+str(i))
    grass.run_command("g.remove", vect="m_pt_3dim"+str(i))
    grass.run_command("g.remove", vect="ptbound"+str(i))
    grass.run_command("g.remove", vect="pt"+str(i))
    grass.run_command("g.remove", rast="cos_aspect"+str(i))
    grass.run_command("g.remove", rast="sen_aspect"+str(i))
    grass.run_command("g.remove", rast="m_pt_3dim"+str(i))
    grass.run_command("g.remove", rast="aspect"+str(i))
    grass.run_command("g.remove", rast="slope"+str(i))
    grass.run_command("g.remove", rast="m_pt_3dim_nn"+str(i))
    grass.run_command("g.remove", rast="asp_semi"+str(i))
    grass.run_command("g.remove", rast="diff_angoli"+str(i))    
    grass.run_command("g.remove", rast="diff_angoli_semi"+str(i))
    grass.run_command("g.remove", rast="diff_angoli_semi_squared"+str(i))
    grass.run_command("g.remove", rast="m_pt_3dim_1_nn_poly")
    grass.run_command("g.remove", rast="m_pt_3dim_nn")
    #incremento il contatore
    k=k+1
		
  #copio il risultato nel layer finale
  grass.run_command("v.category", input="giaciture_mediate", output=out, option="chlayer", type="point", layer="1,1", cat="1", step="1", overwrite=True)	
  grass.run_command("g.region", vect="bedd_3d")
  grass.run_command("g.remove", vect="tmp")
  grass.run_command("g.remove", vect="bed_trc_3d_sinuos")
  grass.run_command("g.remove", vect="bedd_3d")
  grass.run_command("g.remove", vect="connect")
  grass.run_command("g.remove", vect="giaciture_mediate")
  grass.run_command("d.mon", flags="r")  
  return 0
  
if __name__ == "__main__":
  options, flags = grass.parser()
  atexit.register(cleanup)
  sys.exit(main())	

